/*

	mplayerplug-in

	This is a plugin for Netscape and Mozilla that uses mplayer to play supported videoes from websites
	It handles embedded and full page files equally.

	License: GPL

	Created 2003-01-09

	Version 0.80	2003-05-21

	Authors:
	Kevin DeKorte	kdekorte@yahoo.com

 	Contributors
	Fabian Franz	gmplayer patches and support
	Todd Kirby	embed scale aspect option
	Ed Fisher	novop option
	Don Estberg	post processing options and QT Streams
 *
 */

#include <stdio.h>
#include "npapi.h"

// SHM stuff
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>


#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <signal.h>
#include <errno.h>
#include <sys/wait.h>

extern int errno;
int DEBUG;

#define MAX_BUF_LEN 255
#define STATE_RESET 0
#define STATE_NEW 1
#define STATE_HAVEURL 3
#define STATE_WINDOWSET 4
#define STATE_READY 5
#define STATE_PLAYING 10
//#define DEBUG 1




typedef struct _PluginInstance
{
	int state;
	char* url;
	char* fname;
	char* href;
	uint16 mode;
	Widget netscape_widget;
	Display* display;
	Window window;
	uint32 width;
	uint32 height;
	uint32 embed_width;
	uint32 embed_height;
	pid_t mplayer;
	FILE* logfp;
	char* logfile;
	char* output_display;
	int cachesize;
	int use_gmplayer;
	char* player;
	int setwindow;
	int loop;
	int maintain_aspect;
	int scaling;
	int fileonly_embed;
	int rtsp_use_tcp;
	char* vo;
	char* ao;
	char* vopopt;
	int noembed;
	int novop;
	int osdlevel;
	int shmkey1;
	int shmkey2;
	int use_streaming;
	FILE* pipe_fd[2];
	uint32 nQtNext;
	char* qtNext[256];
	int download;
	char* download_dir;
	int forced;
} PluginInstance;



void Redraw(Widget w, NPP instance, XEvent *event);
void CloseEvents(Widget w, XtPointer instance, XAnyEvent *event, Boolean *b);
static void DestroyCB (Widget widget, XtPointer client_data, XtPointer call_data);
static void ResizeCB (Widget widget, XtPointer client_data, XtPointer call_data);
void lowercase(char string[]);

void lowercase(char string[])
{
   int  i = 0;

   while ( string[i] )
   {
      string[i] = tolower(string[i]);
      i++;
   }

   return;
}



char*
NPP_GetMIMEDescription(void)
{
	/* type of files we handle */
	return("video/quicktime:mov:Quicktime;"
		"video/x-quicktime:mov:Quicktime;"
		"video/quicktime:mp4:Quicktime;"
		"video/quicktime:sdp:Quicktime - Session Description Protocol;"
		"application/x-quicktimeplayer:mov:Quicktime;"
		"video/x-msvideo:avi:AVI;"
		"video/msvideo:avi:AVI;"
		"video/mpeg:mpg,mpeg:MPEG;"
		"audio/mpeg:mpg,mpeg:MPEG;"
		"video/x-mpeg:mpg,mpeg:MPEG;"
		"video/x-mpeg2:mpv2,mp2ve:MPEG2;"
		"audio/mpeg:mpg,mpeg:MPEG;"
		"audio/x-mpeg:mpg,mpeg:MPEG;"
		"audio/mpeg2:mp2:MPEG audio;"
		"audio/x-mpeg2:mp2:MPEG audio;"
		"audio/mpeg3:mp3:MPEG audio;"
		"audio/x-mpeg3:mp3:MPEG audio;"
		"audio/mp3:mp3:MPEG audio;"
		"audio/x-pn-realaudio:ram:RealAudio;"
		"audio/x-pn-realaudio-plugin:rpm:RealAudio;"
		"audio/x-realaudio:ra:RealAudio;"
		"video/fli:fli,flc:FLI animation;"
		"video/x-fli:fli,flc:FLI animation;"
		"video/x-ms-asf-plugin:asf,asx:Windows Media;"
		"video/dl:dl:DL animation;"
		"video/x-dl:dl:DL animation;"
		"video/sgi-movie:movie,moci,mv:SGI animation;"
		"video/x-sgi-movie:movie,moci,mv:SGI animation;"
		"video/anim:iff,anim5,anim3,anim7:IFF animation;"
		"video/x-anim:iff,anim5,anim3,anim7:IFF animation;"
		"application/x-ogg:ogg,ogm:Ogg Vorbis Media;"
		"audio/ogg:ogg,ogm:Ogg Vorbis Audio;"
		"application/x-mplayer2:wmv:WMV;"
		"video/x-ms-wm:asf:MSNBCPlayer;"
		"video/x-ms-asf:asf,asx:Windows Media;"
		"video/x-ms-wmv:wmv:Windows Media;"
		"video/x-ms-wvx:wvx:Windows Media;"	
		"audio/x-ms-wax:wax:Windows Media;"
		"application/x-drm-v2:asx:Windows Media;"
		"audio/wav:wav:Microsoft wave file;"
		"audio/x-wav:wav:Microsoft wave file"
	);
}

NPError
NPP_GetValue(void *future, NPPVariable variable, void *value)
{
	NPError err = NPERR_NO_ERROR;

	switch (variable) {
		case NPPVpluginNameString:
			// have to have those names in the name string to make Atom Films work
			*((char **)value) = "mplayerplug-in v0.80 handles QuickTime Windows Media Player Plugin";
			break;
		case NPPVpluginDescriptionString:
			*((char **)value) =
				"Video Player Plug-in for QuickTime and Windows Media Player streams using mplayer";
			break;
		default:
			err = NPERR_GENERIC_ERROR;
	}
	return err;
}

NPError
NPP_Initialize(void)
{
	return NPERR_NO_ERROR;
}


jref
NPP_GetJavaClass()
{
	return NULL;
}

void
NPP_Shutdown(void)
{
}


NPError
NPP_New(NPMIMEType pluginType,
	NPP instance,
	uint16 mode,
	int16 argc,
	char* argn[],
	char* argv[],
	NPSavedData* saved)
{
        PluginInstance* This;
	int i,j;
	FILE* config;
	char buffer[1000];
	char parse[1000];
	char logfile[1000];
	char* cp;

	if (instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;

	instance->pdata = NPN_MemAlloc(sizeof(PluginInstance));

	This = (PluginInstance*) instance->pdata;

	This->state=STATE_NEW;
	This->url=NULL;
	This->fname=NULL;
	This->href=NULL;
	This->mode=mode;
	This->netscape_widget=NULL;
	This->display=NULL;
	This->window=(Window)NULL;
	This->width=0;
	This->height=0;
	This->embed_height=0;
	This->embed_width=0;
	This->mplayer=0;
	This->setwindow=0;
	This->forced=0;
	This->nQtNext=0;
	for (i=0; i < 256; i++)
		This->qtNext[i] = NULL;


	if (DEBUG) printf("NEW %i\n",instance);


	// load config file

	config = NULL;

	if (config == NULL) {
		sprintf(buffer,"%s",getenv("HOME"));
		strcat(buffer,"/.mplayer/mplayerplug-in.conf");
		config = fopen(buffer,"r");
	}

	if (config == NULL) {
		sprintf(buffer,"%s",getenv("HOME"));
		strcat(buffer,"/.mozilla/mplayerplug-in.conf");
		config = fopen(buffer,"r");
	}

	if (config == NULL) {
		config = fopen("/etc/mplayerplug-in.conf","r");
	}

	DEBUG = 0;
	This->cachesize=0;
	This->logfile=NULL;
	This->logfp=NULL;
	This->player=strdup("mplayer");
	This->use_gmplayer=0;
	This->vo=NULL;
	This->ao=NULL;
	This->noembed=0;
	This->novop=0;
	This->osdlevel=0;
	This->vopopt=NULL;
	This->maintain_aspect=1;
	This->scaling=1;
	This->fileonly_embed=1;
	This->rtsp_use_tcp=0;
	This->use_streaming=0;
	This->output_display=NULL;
	This->download=0;
	This->download_dir=NULL;

	if (config == NULL) {
		// no config file
	} else {
		while(fgets(buffer,1000,config) != NULL) {
			if ( strncasecmp(buffer,"cachesize",9) == 0) {
				sprintf(parse,"%s",strtok(buffer,"="));
				sprintf(parse,"%s",strtok(NULL,"="));
				sscanf(parse,"%i",&This->cachesize);
				if (This->cachesize < 0) This->cachesize=0;
				if (This->cachesize > 65535) This->cachesize=65535;
				continue;
			}

			if ( strncasecmp(buffer,"debug",5) == 0) {
				sprintf(parse,"%s",strtok(buffer,"="));
				sprintf(parse,"%s",strtok(NULL,"="));
				sscanf(parse,"%i",&DEBUG);
				if (DEBUG != 0) DEBUG=1;
				continue;
			}

			if ( strncasecmp(buffer,"noembed",7) == 0) {
				sprintf(parse,"%s",strtok(buffer,"="));
				sprintf(parse,"%s",strtok(NULL,"="));
				sscanf(parse,"%i",&This->noembed);
				if (This->noembed != 0) This->noembed=1;
				continue;
			}
			
			if ( strncasecmp(buffer,"novop",5) == 0) {
				sprintf(parse,"%s",strtok(buffer,"="));
				sprintf(parse,"%s",strtok(NULL,"="));
				sscanf(parse,"%i",&This->novop);
				if (This->novop != 0) This->novop=1;
				continue;
			}
			if ( strncasecmp(buffer,"vopopt",6) == 0) {
				sprintf(parse,"%s",strtok(buffer,"="));
				sprintf(parse,"%s",strtok(NULL,"\n"));
				This->novop=0;
				This->vopopt=strdup(parse);
				continue;
			}

			if ( strncasecmp(buffer,"prefer-aspect",13) == 0) {
				sprintf(parse,"%s",strtok(buffer,"="));
				sprintf(parse,"%s",strtok(NULL,"="));
				sscanf(parse,"%i",&This->maintain_aspect);
				if (This->maintain_aspect != 0) This->maintain_aspect=1;
				continue;
			}

			if ( strncasecmp(buffer,"scaling",7) == 0) {
				sprintf(parse,"%s",strtok(buffer,"="));
				sprintf(parse,"%s",strtok(NULL,"="));
				sscanf(parse,"%i",&This->scaling);
				if (This->scaling != 0) This->scaling=1;
				continue;
			}
			if ( strncasecmp(buffer,"fileonly-embed",14) == 0) {
				sprintf(parse,"%s",strtok(buffer,"="));
				sprintf(parse,"%s",strtok(NULL,"="));
				sscanf(parse,"%i",&This->fileonly_embed);
				if (This->fileonly_embed != 0) This->fileonly_embed=1;
				continue;
			}

			if ( strncasecmp(buffer,"rtsp-use-tcp",12) == 0) {
				sprintf(parse,"%s",strtok(buffer,"="));
				sprintf(parse,"%s",strtok(NULL,"="));
				sscanf(parse,"%i",&This->rtsp_use_tcp);
				if (This->rtsp_use_tcp != 0) This->rtsp_use_tcp=1;
				continue;
			}

			if ( strncasecmp(buffer,"use-streaming",13) == 0) {
				sprintf(parse,"%s",strtok(buffer,"="));
				sprintf(parse,"%s",strtok(NULL,"="));
				sscanf(parse,"%i",&This->use_streaming);
				if (This->use_streaming != 0) This->use_streaming=1;
				continue;
			}

			if ( strncasecmp(buffer,"logfile",7) == 0) {
				sprintf(parse,"%s",strtok(buffer,"="));
				sprintf(logfile,"%s",strtok(NULL,"=\n"));
				if (strstr(logfile,"$HOME") != NULL) {
					sprintf(buffer,"%s%s\0",getenv("HOME"),logfile+5);
					strcpy(logfile,buffer);
				}
				This->logfile=strdup(logfile);
				This->logfp = fopen(This->logfile,"a");
				continue;
			}

			if ( strncasecmp(buffer,"use-gui",7) == 0) {
				sprintf(parse,"%s",strtok(buffer,"="));
				sprintf(parse,"%s",strtok(NULL,"=\n"));
				if (strncasecmp(parse, "yes", 3) == 0) {
					This->use_gmplayer=1;
				} else {
					if (strncasecmp(parse, "mini", 4) == 0) {
						This->use_gmplayer=2;
					}
				}

				if (This->use_gmplayer > 0) {
					if (This->player) free(This->player);
					This->player=strdup("gmplayer");
				}
				continue;
			}

			if ( strncasecmp(buffer,"vo",2) == 0) {
				sprintf(parse,"%s",strtok(buffer,"="));
				sprintf(parse,"%s",strtok(NULL,"=\n"));
				This->vo=strdup(parse);
				continue;
			}

			if ( strncasecmp(buffer,"ao",2) == 0) {
				sprintf(parse,"%s",strtok(buffer,"="));
				sprintf(parse,"%s",strtok(NULL,"=\n"));
				This->ao=strdup(parse);
				continue;
			}

			if ( strncasecmp(buffer,"display",7) == 0) {
				sprintf(parse,"%s",strtok(buffer,"="));
				sprintf(parse,"%s",strtok(NULL,"=\n"));
				This->output_display=strdup(parse);
				continue;
			}

			if ( strncasecmp(buffer,"download",8) == 0) {
				sprintf(parse,"%s",strtok(buffer,"="));
				sprintf(parse,"%s",strtok(NULL,"="));
				sscanf(parse,"%i",&This->download);
				if (This->download != 0) This->download=1;
				continue;
			}

			if ( strncasecmp(buffer,"dload-dir",9) == 0) {
				sprintf(parse,"%s",strtok(buffer,"="));
				sprintf(parse,"%s",strtok(NULL,"=\n"));
				if (strstr(parse,"$HOME") != NULL) {
					sprintf(buffer,"%s%s\0",getenv("HOME"),parse+5);
					strcpy(parse,buffer);
				}
				This->download_dir=strdup(parse);
				continue;
			}

			if ( strncasecmp(buffer,"osdlevel",8) == 0) {
				sprintf(parse,"%s",strtok(buffer,"="));
				sprintf(parse,"%s",strtok(NULL,"="));
				sscanf(parse,"%i",&This->osdlevel);
				if (This->osdlevel < 0)
					This->osdlevel = 0;
				if (This->osdlevel > 3)
					This->osdlevel = 3;
			}

		}
	}
	if (This->logfp == NULL) DEBUG=0;

	if (mode == NP_EMBED) {

		if (DEBUG) fprintf(This->logfp,"Embedded mode\n");
		for(i=0;i<argc;i++){
			if (DEBUG) fprintf(This->logfp,"Argument Name: %s\n",argn[i]);
			if (DEBUG) fprintf(This->logfp,"Argument Value: %s\n",argv[i]);

			if((strncasecmp(argn[i],"src",3) == 0) || (strncasecmp(argn[i],"filename",8) == 0)){
				This->url=strdup(argv[i]);
				This->state=STATE_HAVEURL;
				if (DEBUG) fprintf(This->logfp,"\n==%s==%d\n",This->url,i);
			}

			if((strncasecmp(argn[i],"href",4) == 0) || (strncasecmp(argn[i],"qtsrc",5) == 0)){
				This->href=strdup(argv[i]);
				This->state=STATE_HAVEURL;
				if (DEBUG) fprintf(This->logfp,"\n==%s==%d\n",argv[i],i);
			}

			if(strncasecmp(argn[i],"height",6) == 0) {
				sscanf(argv[i],"%i",&This->embed_height);
			}

			if(strncasecmp(argn[i],"width",5) == 0) {
				sscanf(argv[i],"%i",&This->embed_width);
			}

			/* handle 'scale' attribute used by QT Plugin */
			if (strncasecmp(argn[i],"scale",5) == 0) {
				if (strncasecmp(argv[i],"aspect",6) == 0) {
					This->maintain_aspect=1;
				}
			}

			if(strncasecmp(argn[i],"loop",4) == 0) {

				lowercase(argv[i]);
				if (DEBUG) fprintf(This->logfp,"argv[i]=%s\n",argv[i]);

				if(strstr(argv[i],"true") || strstr(argv[i],"yes")) {
					This->loop=1;
				} else {
					This->loop=0;
				}
				if (DEBUG) fprintf(This->logfp,"loop=%i\n",This->loop);
			}

			if(This->nQtNext<256 && (strncasecmp(argn[i],"qtnext",6) == 0) && argv[i][0]=='<') {
				sprintf(parse,"%s",strtok(&argv[i][1],">"));
				if ( (cp=strchr(parse,' ')) && strlen(parse)==cp-parse+1 )
					*cp = (char)NULL;
				This->qtNext[This->nQtNext++] = strdup(parse);
				sprintf(parse,"%s",strtok(NULL,"<"));
				if (strcmp(parse, "T")) {
					if (DEBUG)
						fprintf(This->logfp,"qtNext%i expected \"T\" found \"%s\"\n",This->nQtNext,parse);
					This->nQtNext--;
				} else {
					sprintf(parse,"%s",strtok(NULL,">\n"));
					if (strcmp(parse, "myself")) {
						if (DEBUG)
							fprintf(This->logfp,"qtNext%i expected \"myself\" found \"%s\"\n",This->nQtNext,parse);
						This->nQtNext--;
					} else if (DEBUG)
						fprintf(This->logfp,"qtNext%i=%s\n",This->nQtNext,This->qtNext[This->nQtNext-1]);
				}
			}



		}
	} else {
		if (DEBUG) fprintf(This->logfp,"New, full mode %i\n", This->mode);
	}


	if (This != NULL)
		return NPERR_NO_ERROR;
	else
		return NPERR_OUT_OF_MEMORY_ERROR;
}


NPError
NPP_Destroy(NPP instance, NPSavedData** save)
{


	PluginInstance* This;
	int ret;

	if (DEBUG) printf("destroy %i\n",instance);

	if (instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;

	This = (PluginInstance*) instance->pdata;
	if (DEBUG) fprintf(This->logfp,"\nIn NPP_Destroy\n");

	if (DEBUG) ret = fclose(This->logfp);

	if (This != NULL) {
		if (DEBUG) printf("Freeing memory\n");
		NPN_MemFree(instance->pdata);
		This = NULL;
		instance->pdata = NULL;
	}

	return NPERR_NO_ERROR;

}



NPError
NPP_SetWindow(NPP instance, NPWindow* window)
{
	PluginInstance* This;
	Widget netscape_widget,screen;
	NPSetWindowCallbackStruct* ws;
	int shmid,ret,i;
	pid_t *shmPtr;
	struct shmid_ds stShmId;
	char buffer[1000];

	if (instance == NULL) {
		if (DEBUG) printf("SetWindow Called with invalid instance\n");
		return NPERR_INVALID_INSTANCE_ERROR;
	}

	This = (PluginInstance*) instance->pdata;

	if ((window == NULL) || (window->window == NULL)) {
		return NPERR_NO_ERROR;
	}


	if (DEBUG) fprintf(This->logfp,"Set Window \nNPP %i\n",instance);
	if (DEBUG) fprintf(This->logfp,"This %i\n",This);
	if (DEBUG) fprintf(This->logfp,"This->state %i\n",This->state);

	if (DEBUG) printf("SW %i\n",instance);

	if (This->window != (Window)window->window) {
		if (DEBUG) fprintf(This->logfp, "Hey we got a new window! old: 0x%x    new 0x%x\n",This->window,window->window);
	}

	if (This->state < STATE_WINDOWSET) {
		if (DEBUG) fprintf(This->logfp,"Size: %d %d %x\n",window->x,window->y,window->window);
		if (DEBUG) fprintf(This->logfp,"Size: %dx%d \n",window->width,window->height,window->window);

		ws = (NPSetWindowCallbackStruct*) window->ws_info;

		netscape_widget = XtWindowToWidget((Display*) ws->display, (Window) window->window);

		XtAddCallback(netscape_widget,XtNdestroyCallback, DestroyCB,(XtPointer) This);

		This->display = ws->display;
		This->window = (Window)window->window;

		This->netscape_widget = netscape_widget;
		This->width = window->width;
		This->height = window->height;
		if (This->state == STATE_HAVEURL) {
			// this is needed for some sites
			if (strncasecmp(This->url,"mms",3) == 0 || strncasecmp(This->url,"rtsp",3) == 0) {
				if (DEBUG) fprintf(This->logfp,"Forced NewStream %s\n",This->url);
				This->state = STATE_READY;
				This->download = 0;
				Redraw(This->netscape_widget, instance, NULL);
			} else {

				if (strncasecmp(This->url,"file://",7) == 0) {

					for (i=7; i < strlen(This->url); i++) {
						buffer[i-7] = This->url[i];
					}
					buffer[i] = '\0';
					This->url = strdup(buffer);
					This->state = STATE_READY;
					This->download = 0;
					Redraw(This->netscape_widget, instance, NULL);

				} else {

					if (This->href == NULL) {
						if (DEBUG) fprintf(This->logfp,"Forced NewStream (URL) %s\n",This->url);
						NPN_GetURL(instance,This->url,NULL);
						This->state = STATE_WINDOWSET;
					} else {
						if (DEBUG) fprintf(This->logfp,"Forced NewStream (HREF) %s\n",This->href);
						This->url=strdup(This->href);
						NPN_GetURL(instance,This->href,NULL);
						free(This->href);
						This->href = NULL;
						This->state = STATE_WINDOWSET;
						This->forced = 1;
					}
				}
			}
		}


	} else {
		if (This->setwindow ==0) {
			if (DEBUG) fprintf(This->logfp,"back in SetWindow\n");
			if (DEBUG) fprintf(This->logfp,"New Size: %ix%i\n",window->width,window->height);
			if (This->mode == NP_EMBED){
				This->embed_width = window->width;
				This->embed_height = window->height;
			}
			This->setwindow++;
		}
	}

	if (DEBUG) fprintf(This->logfp,"Exiting SetWindow\n");
	if (DEBUG) fflush(This->logfp);

	return NPERR_NO_ERROR;
}

void NPP_URLNotify(NPP instance, const char* url, NPReason reason, void* notifyData)
{
	PluginInstance* This;

	if (DEBUG) fprintf(This->logfp,"URL Notify\n");

	if (instance == NULL)
		return;

	This = (PluginInstance*) instance->pdata;

	if (reason == NPRES_NETWORK_ERR) {
		Redraw(This->netscape_widget, instance, NULL);
	}
}

NPError
NPP_NewStream(NPP instance,
			  NPMIMEType type,
			  NPStream *stream,
			  NPBool seekable,
			  uint16 *stype)
{
	NPByteRange range;
	PluginInstance* This;

	if (instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;

	This = (PluginInstance*) instance->pdata;

	if (DEBUG) printf("NS %i\n",instance);
	if (DEBUG) fflush(stdout);

	if (DEBUG) fprintf(This->logfp,"NewStream Requested\n");
	if (DEBUG) fprintf(This->logfp,"State %i Mode:%i\n",This->state, This->mode);
	if (DEBUG) fprintf(This->logfp,"This Window: 0x%x \n",This->window);
	if (DEBUG) fprintf(This->logfp,"URL: %s\n",This->url);
	if (DEBUG) fflush(This->logfp);
	if (DEBUG) printf("NS2 %i\n",instance);
	if (DEBUG) fflush(stdout);

	if (This->mode == NP_FULL) {
		if (DEBUG) fprintf(This->logfp,"NP_ASFILEONLY\n");
		*stype = NP_ASFILEONLY;
	} else {
		if (DEBUG) fprintf(This->logfp,"NP_NORMAL\n");
		*stype = NP_NORMAL;
		if (This->use_streaming == 1) {
			if (DEBUG) fprintf(This->logfp,"Streaming\n");
			if (pipe(This->pipe_fd) < 0) {
				if (DEBUG) fprintf(This->logfp,"pipe failed\n");
			} else {
				if (DEBUG) fprintf(This->logfp,"pipe opened\n");
			}
		}
	}
	if (DEBUG) printf("NS3 %i\n",instance);
	if (DEBUG) fflush(stdout);

	return NPERR_NO_ERROR;

}


/* PLUGIN DEVELOPERS:
 *	These next 2 functions are directly relevant in a plug-in which
 *	handles the data in a streaming manner. If you want zero bytes
 *	because no buffer space is YET available, return 0. As long as
 *	the stream has not been written to the plugin, Navigator will
 *	continue trying to send bytes.  If the plugin doesn't want them,
 *	just return some large number from NPP_WriteReady(), and
 *	ignore them in NPP_Write().  For a NP_ASFILE stream, they are
 *	still called but can safely be ignored using this strategy.
 */

int32 STREAMBUFSIZE = 0X0FFFFFFF; /* If we are reading from a file in NPAsFile
				   * mode so we can take any size stream in our
				   * write call (since we ignore it) */
int32
NPP_WriteReady(NPP instance, NPStream *stream)
{
	PluginInstance* This;
	if (instance != NULL)
		This = (PluginInstance*) instance->pdata;

	if (DEBUG) printf("WR %i\n",instance);

	if (DEBUG) fprintf(This->logfp,"in WriteReady. stream->url: %s\n",stream->url);
	if (DEBUG) fprintf(This->logfp,"in WriteReady, state = %i\n",This->state);
	if (DEBUG) fflush(This->logfp);

	if (DEBUG) fprintf(This->logfp,"Setting of This->forced = %i\n",This->forced);
	if (This->forced == 0) {
		if (This->url) free(This->url);
		This->url = strdup(stream->url);
	}

	if (This->mode == NP_FULL) {
		return STREAMBUFSIZE;
	} else {
		if (This->state >= STATE_READY) {
			if (This->use_streaming == 1) {
				return 4096;
			} else {
				if (DEBUG) printf("destroy stream called\n");
				NPN_DestroyStream(instance,stream,NPRES_DONE);
				return -1;
			}
		} else {
			return 4096;
		}
	}
}


int32
NPP_Write(NPP instance, NPStream *stream, int32 offset, int32 len, void *buffer)
{
	PluginInstance* This;
	int ret;

	if (instance != NULL)
	{
		This = (PluginInstance*) instance->pdata;

	}
	if (DEBUG) printf("W %i\n",instance);
	if (DEBUG) fprintf(This->logfp,"in Write. stream->url: %s\n",stream->url);
	if (DEBUG) fprintf(This->logfp,"in Write, state = %i\n",This->state);
	if (DEBUG) fflush(This->logfp);

	if (This->state >= STATE_READY) {
		if (This->use_streaming == 0) {
			return -1;	// this should kill netscapes stream, the plugin parent is done with the stream, but mplayer has it
		} else {
			ret = write(This->pipe_fd[1],buffer,len);
			if (DEBUG) printf("write %i,%i,%s\n",ret,errno,strerror(errno));
			return len;
		}
	} else {
		This->state=STATE_READY;
		if (This->use_streaming == 1) {
			ret = write(This->pipe_fd[1],buffer,len);
			if (DEBUG) printf("write %i,%i,%s\n",ret,errno,strerror(errno));
		}
		Redraw(This->netscape_widget, instance, NULL);
		return len;
	}

}


NPError
NPP_DestroyStream(NPP instance, NPStream *stream, NPError reason)
{
	PluginInstance* This;
	int ret;

	if (instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;
	This = (PluginInstance*) instance->pdata;

	if (DEBUG) fprintf(This->logfp,"in destroy stream\n");
	if (DEBUG) fflush(This->logfp);
	if (This->use_streaming == 1) {
		ret = close(This->pipe_fd[1]);
		if (DEBUG) printf("close %i,%i,%s\n",ret,errno,strerror(errno));
	}
	return NPERR_NO_ERROR;
}


void
NPP_StreamAsFile(NPP instance, NPStream *stream, const char* fname)
{
	PluginInstance* This;
	FILE *infile;
	char c;
	char s[MAX_BUF_LEN];
	int i=0;

	if (instance != NULL)
	{
		This = (PluginInstance*) instance->pdata;
	}

	if (DEBUG) printf("SaF %i\n",instance);


	if (DEBUG) fprintf(This->logfp,"in stream as file. stream->url: %s\n",stream->url);
	if (DEBUG) fprintf(This->logfp,"in stream as file. fname: %s\n",fname);

	if (This->url) free(This->url);
	This->url = strdup(stream->url);
	if (This->fname) free(This->fname);
	This->fname = strdup(fname);

	if (DEBUG) fprintf(This->logfp,"set state\n");
	This->state=STATE_READY;

	Redraw(This->netscape_widget, instance, NULL);
	return;

}


void
NPP_Print(NPP instance, NPPrint* printInfo)
{
	return;
}



/***********************************************************************
 * Sometimes the plugin widget gets stupidly destroyed, that is whenever
 * Netscape relayouts the page. This callback reparents the client
 * windows to the root window so they do not get destroyed as well.
 * Eventually the NPP_SetWindow function should be called and we'll
 * reparent them back under the plugin.
 ***********************************************************************/
static void
DestroyCB (Widget widget, XtPointer client_data, XtPointer call_data)
{
	PluginInstance* This = (PluginInstance*)client_data;
	int i,ret;
	int shmid;
	pid_t *shmPtr;
	struct shmid_ds stShmId;

	if (DEBUG) fprintf (This->logfp, "DestroyCB, This: %i  Call Data: %i\n", This,call_data);

	if (This->state < STATE_READY) {
		if (DEBUG) fprintf(This->logfp,"State is incorrect %i\n",This->state);
		return;
	}

	if (DEBUG) fflush(This->logfp);

	if (widget == This->netscape_widget ) {
		This->netscape_widget = NULL;

		if (DEBUG) fprintf(This->logfp,"!!!!  Shutdown \n");
		if (DEBUG) fflush(This->logfp);

		// kill the mplayer thread
		shmid = shmget(This->shmkey2,sizeof(pid_t),0);
		shmPtr = shmat(shmid,0,0);
		if (*shmPtr != 0) {
			if (DEBUG) fprintf(This->logfp,"Sending PID %i a SIGTERM signal\n",*shmPtr);
			ret = kill(*shmPtr,SIGTERM);
			if (DEBUG) fprintf(This->logfp,"Mplayer Kill return %i:%i:%s\n",ret,errno,strerror(errno));
		}
		shmdt(shmPtr);


		shmid = shmget(This->shmkey1,sizeof(pid_t),0);
		shmPtr = shmat(shmid,0,0);
		if (*shmPtr != 0)  {
			if (DEBUG) fprintf(This->logfp,"Sending PID %i a SIGINT signal\n",*shmPtr);
			ret = kill(*shmPtr,SIGINT);
			if (DEBUG) fprintf(This->logfp,"Child Kill return %i:%i:%s\n",ret,errno,strerror(errno));
		}
		shmdt(shmPtr);

		// cleanup shared ram
		shmid = shmget(This->shmkey1,sizeof(pid_t),0);
		ret = shmctl(shmid,IPC_RMID,&stShmId);
		if (DEBUG) fprintf(This->logfp,"Shm key %i remove return %i:%i:%s\n",This->shmkey1,ret,errno,strerror(errno));

		shmid = shmget(This->shmkey2,sizeof(pid_t),0);
		ret = shmctl(shmid,IPC_RMID,&stShmId);
		if (DEBUG) fprintf(This->logfp,"Shm key %i remove return %i:%i:%s\n",This->shmkey2,ret,errno,strerror(errno));

	} else {
		if (DEBUG) fprintf(This->logfp,"Widgets don't match %i != %i\n",widget,This->netscape_widget);


	}
	if (DEBUG) fflush(This->logfp);

}




void Redraw(Widget w, NPP instance, XEvent *event)
{


	GC gc;
        int pid,ret, i, j,k;
        pid_t child, child2, mp_pid, moz_pid;
        char wval[32], xval[32], yval[32], yval2[32], video[1000], cacheval[1000], vop[120], cmd[1024], local[1024], osdval[32];
	char *argv[30], *buffer, *fname;

        XGCValues gcv;
        const char* textload = "mplayerplug-in - Loading Movie..";
        const char* textdl = "mplayerplug-in - Downloading Movie..";
	const char* complete = "mplayer has finished playing the movie";
	int status;
	struct sigaction action;
	int shmid,shmid2;
	pid_t *shmPtr;
	PluginInstance* This;
	FILE* saved;

	if (DEBUG) printf("RED %i\n",instance);

	if (instance == NULL) {
		if (DEBUG) printf("Our instance is null!\n");
		return;
	}

        This = (PluginInstance*)instance->pdata;

	if (This == NULL) {
		if (DEBUG) fprintf(This->logfp,"Our This is null!\n");
		if (DEBUG) fflush(This->logfp);
		return;
	}

	if (DEBUG) fprintf(This->logfp,"getting ready to fork mplayer\n");
	if (DEBUG) fprintf(This->logfp,"State: %i   Event: %i \n",This->state,event);
	if (DEBUG) fflush(This->logfp);

	if (This->state != STATE_READY) {
		if (DEBUG) fprintf(This->logfp,"Not ready to fork\n");
		if (DEBUG) fflush(This->logfp);
		return;
	}

	if ( w == NULL) {
		if (DEBUG) fprintf(This->logfp,"Our widget is null!\n");
		if (DEBUG) fflush(This->logfp);
		return;
	}

	if ( w != This->netscape_widget) {
		if (DEBUG) fprintf(This->logfp,"Our widgets don't match!\n");
		if (DEBUG) fflush(This->logfp);
		return;
	}

	if (DEBUG) fprintf(This->logfp,"Window -- 0x%x\n",This->window);
	if (DEBUG) fprintf(This->logfp,"URL -- %s\n",This->url);
	if (DEBUG) fprintf(This->logfp,"File -- %s\n",This->fname);
	if (DEBUG) fprintf(This->logfp,"Size %ix%i\n",This->width,This->height);
	if (DEBUG) fprintf(This->logfp,"Embed Size %ix%i\n",This->embed_width,This->embed_height);
	if (DEBUG) fflush(This->logfp);


	XtVaGetValues(w, XtNbackground, &gcv.background,XtNforeground, &gcv.foreground, 0);
	gc = XCreateGC((Display*)This->display, This->window, GCForeground|GCBackground, &gcv);
	XDrawRectangle((Display*)This->display, This->window, gc, 0, 0, This->width-1, This->height-1);
	if (This->download == 1) {
		XDrawString((Display*)This->display, This->window, gc, 10, 20,textdl, strlen(textdl));
	} else {
		XDrawString((Display*)This->display, This->window, gc, 10, 20,textload, strlen(textload));
	}
	XFreeGC((Display *)This->display,gc);



	if (DEBUG) fprintf(This->logfp,"\n\n ---------Forking------------ \n\n");
	moz_pid=getpid();
	if (DEBUG) fprintf(This->logfp,"\nMoz Pid %i\n",moz_pid);

	This->shmkey1 = (int)random();
	This->shmkey2 = (int)random();

	shmid = shmget(This->shmkey1,sizeof(pid_t),0600 | IPC_CREAT);
	shmid2 = shmget(This->shmkey2,sizeof(pid_t),0600 | IPC_CREAT);

	child=fork();
	setsid();
	setpgid(child,moz_pid);

	if(child==0)
	{

		if (DEBUG) fprintf(This->logfp,"\nsucessfully forked %i:%i\n",getpid(),getpgid(getpid()));
		if(strstr(This->url,"'")){
			if (DEBUG) fprintf(This->logfp,"\n-------- ' is not allowed ------\n");
		} else {
			if (DEBUG) fprintf(This->logfp,"\n-------- url seems to be ok ---\n");

			sprintf(wval,"0x%x",This->window);
			sprintf(cacheval,"%i",This->cachesize);
			sprintf(osdval,"%i",This->osdlevel);

			if (This->mode == NP_FULL) {
				sprintf(video,"%s",This->fname);
				sprintf(xval,"%i",This->width);
				sprintf(yval,"%i",This->height);
				sprintf(yval2,"%i",This->height-16);
				if(This->fileonly_embed==0) This->noembed=1;
			} else {
				if (This->href) {
					sprintf(video,"%s",This->href);
				} else {
					sprintf(video,"%s",This->url);
				}
				sprintf(xval,"%i",This->embed_width);
				sprintf(yval,"%i",This->embed_height);
				sprintf(yval2,"%i",This->embed_height-16);

			}

			// if you don't fork again then the browser waits for the video to be done.
			// Browser is in a "WAIT" state
			child2=fork();
			This->mplayer=child2;
			if(child2==0){
				setsid();
		 		setpgid(child2,child);

				if (This->use_streaming == 1 && This->mode == NP_EMBED) {
					sprintf(video,"-");
					close(0);
					ret = dup2(This->pipe_fd[0],0);
					if (DEBUG) printf("dup2 %i,%i,%s\n",ret,errno,strerror(errno));
					close(This->pipe_fd[0]);
					if (This->cachesize < 32) {
						This->cachesize=32;
						sprintf(cacheval,"%i",This->cachesize);
					}
				}

				if ( XtWindow(w) != This->window ) {
					if (DEBUG) printf("Something wrong with the window\n");
				} else {

					i = 0;
					argv[i++] = strdup(This->player);
					if (DEBUG == 0) {
						argv[i++] = "-really-quiet";
					} else {
						argv[i++] = "-quiet";
					}

					if (This->novop == 1) {
						argv[i++] = "-vop";
						argv[i++] = "null";
					} else {
						if (This->vopopt != NULL) {
							argv[i++] = "-vop";
							argv[i++] = strdup(This->vopopt);
							if (strncmp(This->vopopt,"scale=",6)==0) {
								argv[i++] = "-fs";
							}
						}
					}

					if (This->noembed == 0) {
						argv[i++] = "-wid";
						argv[i++] = wval;
					}

					if ( (This->embed_width == 0) && (This->mode == NP_EMBED)) {

					} else {
						if(This->scaling == 1) {
							if (This->use_gmplayer > 0) {
								argv[i++] = "-screenw";
								argv[i++] = xval;
								argv[i++] = "-screenh";
								argv[i++] = yval;
								if (This->maintain_aspect == 1) {
									argv[i++] = "-xy";
									argv[i++] = xval;
								} else {
									argv[i++] = "-x";
									argv[i++] = xval;
								}
								if (This->use_gmplayer == 2) {
									if (This->maintain_aspect == 0) {
										argv[i++] = "-y";
										argv[i++] = yval2;
									}
									argv[i++] = "-skin";
									argv[i++] = "mini";
									if (This->noembed == 0) {
										argv[i++] = "-guiwid";
										argv[i++] = wval;
									}
								} else {
									if (This->maintain_aspect == 0) {
										argv[i++] = "-y";
										argv[i++] = yval;
									}
								}
							} else {
								if (This->maintain_aspect == 1) {
									argv[i++] = "-xy";
									argv[i++] = xval;
								} else {
									argv[i++] = "-x";
									argv[i++] = xval;
									argv[i++] = "-y";
									argv[i++] = yval;
								}
							}
						}

					}

					if (This->rtsp_use_tcp == 1) {
						argv[i++] = "-rtsp-stream-over-tcp";
					}

					if (This->cachesize > 0 ) {
						argv[i++] = "-cache";
						argv[i++] = cacheval;
					}

					if (This->loop == 1) {
						argv[i++] = "-loop";
						argv[i++] = "0";
					}

					if (This->vo) {
						argv[i++] = "-vo";
						argv[i++] = strdup(This->vo);
					}

					if (This->ao) {
						argv[i++] = "-ao";
						argv[i++] = strdup(This->ao);
					}

					if (This->output_display) {
						argv[i++] = "-display";
						argv[i++] = strdup(This->output_display);
					}

					argv[i++] = "-osdlevel";
					argv[i++] = osdval;

					/* some extra flags for mplayer */
					argv[i++] = "-zoom"; /* -vo x11 needs -zoom for rescaling */
					argv[i++] = "-nojoystick"; /* annoying... */
					argv[i++] = "-nofs"; /* no full screen */

					if (DEBUG) {
						printf("nQtNext=%i\n",This->nQtNext);
						k=0;
						while (This->qtNext[k] != NULL) {
							if (DEBUG) printf("qtNext[%i]=%s\n",k,This->qtNext[k++]);
							fflush(stdout);
						}
					}

					argv[i++] = video;
					j=0;
					while (This->qtNext[j] != NULL) {
						argv[i++] = strdup(This->qtNext[j++]);
					}
					argv[i++] = NULL;


					if (This->download == 1) {
						if (DEBUG) {
							printf("ready to parse\n");
							fflush(stdout);
						}
						// get the filename from the end of the URL
						buffer = strtok(This->url,"/");
						while(buffer != NULL) {
							fname = strdup(buffer);
							if (DEBUG) printf("parsed: %s\n",fname);
							buffer = strtok(NULL,"/");
						}
						if (DEBUG) printf("filename: %s\n",fname);
						//have we downloaded it before
						sprintf(local,"%s/%s",This->download_dir,fname);
						if  (DEBUG) printf("local filename: %s\n",local);
						saved = fopen(local,"r");
						if (DEBUG) printf("File pointer %i\n",saved);
						if ( saved == NULL) {
							// no
							if (This->mode == NP_EMBED) {
								sprintf(cmd,"mplayer -dumpstream -dumpfile \"%s/%s\" `mplayer -v -nosound -novideo \"%s\" | grep Playing | tail -n 1 | awk '{print $2}'`",This->download_dir,fname,video);
								if (DEBUG) {
									printf("%s\n",cmd);
									fflush(stdout);
								}
								ret = system(cmd);
							} else {
								sprintf(cmd,"cp \"%s\" \"%s/%s\"",video,This->download_dir,fname);
								ret = system(cmd);
							}
						} else {
							//yes
							fclose(saved);
						}

						sprintf(argv[i-2],"%s/%s",This->download_dir,fname);
						if (DEBUG) {
							for (i=0; argv[i] != NULL; i++) {
								printf("argv[%i]=%s\n",i,argv[i]);
							}
							printf("%s\n",This->href);
							printf("%s\n",This->url);

						}
						ret = execvp(argv[0],argv);
					} else {
						if (DEBUG) {
							for (i=0; argv[i] != NULL; i++) {
								printf("argv[%i]=%s\n",i,argv[i]);
							}
							printf("%s\n",This->href);
							printf("%s\n",This->url);

						}
						ret = execvp(argv[0],argv);
					}

				}


			} else {
				if(child2 < 0)
					abort();

				if (This->use_streaming == 1) {
					close(This->pipe_fd[0]);
				}
				// use shared memory to pass the pid back to the parent process
				shmid2 = shmget(This->shmkey2,sizeof(pid_t),0);
				shmPtr = shmat(shmid2,0,0);
				*shmPtr = child2;
				shmdt(shmPtr);

			}
			_exit(0);
		}
		_exit(0);
	} else {
		if (child < 0)
			abort();
		else  {
			if (DEBUG) fprintf(This->logfp,"\n---- forked PID of mplayer (child) %d ----\n",child);
			if (This->use_streaming == 1) {
				close(This->pipe_fd[0]);
			}
			NPN_Status(instance,"Done");
			shmPtr = shmat(shmid,0,0);
			*shmPtr = child;
			if (DEBUG) fflush(This->logfp);
			wait(0);
			if (DEBUG) fprintf(This->logfp,"\n---- Woke up from wait ----\n");
			shmdt(shmPtr);
			if (DEBUG) fflush(This->logfp);
			return;
		}
	}
}


